@@ -1,18 +1,18 @@
-/*  Lzip - LZMA lossless data compressor
-    Copyright (C) 2008-2014 Antonio Diaz Diaz.
+/* Lzip - LZMA lossless data compressor
+   Copyright (C) 2008-2024 Antonio Diaz Diaz.
 
-    This program is free software: you can redistribute it and/or modify
-    it under the terms of the GNU General Public License as published by
-    the Free Software Foundation, either version 2 of the License, or
-    (at your option) any later version.
+   This program is free software: you can redistribute it and/or modify
+   it under the terms of the GNU General Public License as published by
+   the Free Software Foundation, either version 2 of the License, or
+   (at your option) any later version.
 
-    This program is distributed in the hope that it will be useful,
-    but WITHOUT ANY WARRANTY; without even the implied warranty of
-    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-    GNU General Public License for more details.
+   This program is distributed in the hope that it will be useful,
+   but WITHOUT ANY WARRANTY; without even the implied warranty of
+   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+   GNU General Public License for more details.
 
-    You should have received a copy of the GNU General Public License
-    along with this program.  If not, see <http://www.gnu.org/licenses/>.
+   You should have received a copy of the GNU General Public License
+   along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
 
 #define _FILE_OFFSET_BITS 64
@@ -26,123 +26,26 @@
 #include <stdint.h>
 
 #include "lzip.h"
+#include "encoder_base.h"
 #include "encoder.h"
 
 
-Dis_slots dis_slots;
-Prob_prices prob_prices;
+const CRC32 crc32;
 
 
-bool Matchfinder_base::read_block()
+int LZ_encoder::get_match_pairs( Pair * pairs )
   {
-  if( !at_stream_end && stream_pos < buffer_size )
-    {
-    const int size = buffer_size - stream_pos;
-    const int rd = readblock( infd, buffer + stream_pos, size );
-    stream_pos += rd;
-    if( rd != size && errno ) throw Error( "Read error" );
-    if( rd < size ) { at_stream_end = true; pos_limit = buffer_size; }
-    }
-  return pos < stream_pos;
-  }
-
-
-void Matchfinder_base::normalize_pos()
-  {
-  if( pos > stream_pos )
-    internal_error( "pos > stream_pos in Matchfinder_base::normalize_pos." );
-  if( !at_stream_end )
-    {
-    const int offset = pos - dictionary_size_ - before_size;
-    const int size = stream_pos - offset;
-    std::memmove( buffer, buffer + offset, size );
-    partial_data_pos += offset;
-    pos -= offset;
-    stream_pos -= offset;
-    for( int i = 0; i < num_prev_positions; ++i )
-      prev_positions[i] -= std::min( prev_positions[i], offset );
-    for( int i = 0; i < pos_array_size; ++i )
-      pos_array[i] -= std::min( pos_array[i], offset );
-    read_block();
-    }
-  }
-
-
-Matchfinder_base::Matchfinder_base( const int before, const int dict_size,
-                    const int after_size, const int dict_factor,
-                    const int num_prev_positions23,
-                    const int pos_array_factor, const int ifd )
-  :
-  partial_data_pos( 0 ),
-  before_size( before ),
-  pos( 0 ),
-  cyclic_pos( 0 ),
-  stream_pos( 0 ),
-  infd( ifd ),
-  at_stream_end( false )
-  {
-  const int buffer_size_limit =
-    ( dict_factor * dict_size ) + before_size + after_size;
-  buffer_size = std::max( 65536, dict_size );
-  buffer = (uint8_t *)std::malloc( buffer_size );
-  if( !buffer ) throw std::bad_alloc();
-  if( read_block() && !at_stream_end && buffer_size < buffer_size_limit )
-    {
-    buffer_size = buffer_size_limit;
-    uint8_t * const tmp = (uint8_t *)std::realloc( buffer, buffer_size );
-    if( !tmp ) { std::free( buffer ); throw std::bad_alloc(); }
-    buffer = tmp;
-    read_block();
-    }
-  if( at_stream_end && stream_pos < dict_size )
-    dictionary_size_ = std::max( (int)min_dictionary_size, stream_pos );
-  else
-    dictionary_size_ = dict_size;
-  pos_limit = buffer_size;
-  if( !at_stream_end ) pos_limit -= after_size;
-  unsigned size = 1 << std::max( 16, real_bits( dictionary_size_ - 1 ) - 2 );
-  if( dictionary_size_ > 1 << 26 )		// 64 MiB
-    size >>= 1;
-  key4_mask = size - 1;
-  size += num_prev_positions23;
-
-  num_prev_positions = size;
-  pos_array_size = pos_array_factor * ( dictionary_size_ + 1 );
-  size += pos_array_size;
-  prev_positions = new( std::nothrow ) int32_t[size];
-  if( !prev_positions ) { std::free( buffer ); throw std::bad_alloc(); }
-  pos_array = prev_positions + num_prev_positions;
-  for( int i = 0; i < num_prev_positions; ++i ) prev_positions[i] = 0;
-  }
-
-
-void Matchfinder_base::reset()
-  {
-  if( stream_pos > pos )
-    std::memmove( buffer, buffer + pos, stream_pos - pos );
-  partial_data_pos = 0;
-  stream_pos -= pos;
-  pos = 0;
-  cyclic_pos = 0;
-  for( int i = 0; i < num_prev_positions; ++i ) prev_positions[i] = 0;
-  read_block();
-  }
-
-
-int Matchfinder::get_match_pairs( Pair * pairs )
-  {
-  int len_limit = match_len_limit_;
+  int len_limit = match_len_limit;
   if( len_limit > available_bytes() )
     {
     len_limit = available_bytes();
     if( len_limit < 4 ) return 0;
     }
 
-  int maxlen = 0;
+  int maxlen = 3;			// only used if pairs != 0
   int num_pairs = 0;
-  const int pos1 = pos + 1;
-  const int min_pos = ( pos > dictionary_size_ ) ? pos - dictionary_size_ : 0;
-  const uint8_t * const data = buffer + pos;
+  const int min_pos = ( pos > dictionary_size ) ? pos - dictionary_size : 0;
+  const uint8_t * const data = ptr_to_current_pos();
 
   unsigned tmp = crc32[data[0]] ^ data[1];
   const int key2 = tmp & ( num_prev_positions2 - 1 );
@@ -153,35 +56,34 @@ int Matchfinder::get_match_pairs( Pair * pairs )
 
   if( pairs )
     {
-    int np2 = prev_positions[key2];
-    int np3 = prev_positions[key3];
+    const int np2 = prev_positions[key2];
+    const int np3 = prev_positions[key3];
     if( np2 > min_pos && buffer[np2-1] == data[0] )
       {
       pairs[0].dis = pos - np2;
-      pairs[0].len = maxlen = 2;
+      pairs[0].len = maxlen = 2 + ( np2 == np3 );
       num_pairs = 1;
       }
     if( np2 != np3 && np3 > min_pos && buffer[np3-1] == data[0] )
       {
       maxlen = 3;
-      np2 = np3;
-      pairs[num_pairs].dis = pos - np2;
-      ++num_pairs;
+      pairs[num_pairs++].dis = pos - np3;
       }
     if( num_pairs > 0 )
       {
-      const int delta = pos1 - np2;
+      const int delta = pairs[num_pairs-1].dis + 1;
       while( maxlen < len_limit && data[maxlen-delta] == data[maxlen] )
         ++maxlen;
       pairs[num_pairs-1].len = maxlen;
+      if( maxlen < 3 ) maxlen = 3;
       if( maxlen >= len_limit ) pairs = 0;	// done. now just skip
       }
-    if( maxlen < 3 ) maxlen = 3;
     }
 
+  const int pos1 = pos + 1;
   prev_positions[key2] = pos1;
   prev_positions[key3] = pos1;
-  int newpos = prev_positions[key4];
+  int newpos1 = prev_positions[key4];
   prev_positions[key4] = pos1;
 
   int32_t * ptr0 = pos_array + ( cyclic_pos << 1 );
@@ -190,12 +92,12 @@ int Matchfinder::get_match_pairs( Pair * pairs )
 
   for( int count = cycles; ; )
     {
-    if( newpos <= min_pos || --count < 0 ) { *ptr0 = *ptr1 = 0; break; }
+    if( newpos1 <= min_pos || --count < 0 ) { *ptr0 = *ptr1 = 0; break; }
 
-    const int delta = pos1 - newpos;
+    const int delta = pos1 - newpos1;
     int32_t * const newptr = pos_array +
       ( ( cyclic_pos - delta +
-          ( ( cyclic_pos >= delta ) ? 0 : dictionary_size_ + 1 ) ) << 1 );
+          ( ( cyclic_pos >= delta ) ? 0 : dictionary_size + 1 ) ) << 1 );
     if( data[len-delta] == data[len] )
       {
       while( ++len < len_limit && data[len-delta] == data[len] ) {}
@@ -214,16 +116,16 @@ int Matchfinder::get_match_pairs( Pair * pairs )
       }
     if( data[len-delta] < data[len] )
       {
-      *ptr0 = newpos;
+      *ptr0 = newpos1;
       ptr0 = newptr + 1;
-      newpos = *ptr0;
+      newpos1 = *ptr0;
       len0 = len; if( len1 < len ) len = len1;
       }
     else
       {
-      *ptr1 = newpos;
+      *ptr1 = newpos1;
       ptr1 = newptr;
-      newpos = *ptr1;
+      newpos1 = *ptr1;
       len1 = len; if( len0 < len ) len = len0;
       }
     }
@@ -231,38 +133,6 @@ int Matchfinder::get_match_pairs( Pair * pairs )
   }
 
 
-void Range_encoder::flush_data()
-  {
-  if( pos > 0 )
-    {
-    if( outfd >= 0 && writeblock( outfd, buffer, pos ) != pos )
-      throw Error( "Write error" );
-    partial_member_pos += pos;
-    pos = 0;
-    if( verbosity >= 2 ) show_progress();
-    }
-  }
-
-
-     // End Of Stream mark => (dis == 0xFFFFFFFFU, len == min_match_len)
-void LZ_encoder_base::full_flush( const unsigned long long data_position,
-                                  const State state )
-  {
-  const int pos_state = data_position & pos_state_mask;
-  renc.encode_bit( bm_match[state()][pos_state], 1 );
-  renc.encode_bit( bm_rep[state()], 0 );
-  encode_pair( 0xFFFFFFFFU, min_match_len, pos_state );
-  renc.flush();
-  File_trailer trailer;
-  trailer.data_crc( crc() );
-  trailer.data_size( data_position );
-  trailer.member_size( renc.member_position() + File_trailer::size() );
-  for( int i = 0; i < File_trailer::size(); ++i )
-    renc.put_byte( trailer.data[i] );
-  renc.flush_data();
-  }
-
-
 void LZ_encoder::update_distance_prices()
   {
   for( int dis = start_dis_model; dis < modeled_distances; ++dis )
@@ -270,7 +140,7 @@ void LZ_encoder::update_distance_prices()
     const int dis_slot = dis_slots[dis];
     const int direct_bits = ( dis_slot >> 1 ) - 1;
     const int base = ( 2 | ( dis_slot & 1 ) ) << direct_bits;
-    const int price = price_symbol_reversed( bm_dis + base - dis_slot - 1,
+    const int price = price_symbol_reversed( bm_dis + ( base - dis_slot ),
                                              dis - base, direct_bits );
     for( int len_state = 0; len_state < len_states; ++len_state )
       dis_prices[len_state][dis] = price;
@@ -297,9 +167,9 @@ void LZ_encoder::update_distance_prices()
   }
 
 
-/* Return value == number of bytes advanced (ahead).
+/* Return the number of bytes advanced (ahead).
    trials[0]..trials[ahead-1] contain the steps to encode.
-   ( trials[0].dis == -1 ) means literal.
+   ( trials[0].dis4 == -1 ) means literal.
    A match/rep longer or equal than match_len_limit finishes the sequence.
 */
 int LZ_encoder::sequence_optimizer( const int reps[num_rep_distances],
@@ -320,37 +190,36 @@ int LZ_encoder::sequence_optimizer( const int reps[num_rep_distances],
   int rep_index = 0;
   for( int i = 0; i < num_rep_distances; ++i )
     {
-    replens[i] = matchfinder.true_match_len( 0, reps[i] + 1, max_match_len );
+    replens[i] = true_match_len( 0, reps[i] + 1 );
     if( replens[i] > replens[rep_index] ) rep_index = i;
     }
-  if( replens[rep_index] >= matchfinder.match_len_limit() )
+  if( replens[rep_index] >= match_len_limit )
     {
-    trials[0].dis = rep_index;
     trials[0].price = replens[rep_index];
-    move_pos( replens[rep_index] );
+    trials[0].dis4 = rep_index;
+    move_and_update( replens[rep_index] );
     return replens[rep_index];
     }
 
-  if( main_len >= matchfinder.match_len_limit() )
+  if( main_len >= match_len_limit )
     {
-    trials[0].dis = pairs[num_pairs-1].dis + num_rep_distances;
     trials[0].price = main_len;
-    move_pos( main_len );
+    trials[0].dis4 = pairs[num_pairs-1].dis + num_rep_distances;
+    move_and_update( main_len );
     return main_len;
     }
 
-  const int pos_state = matchfinder.data_position() & pos_state_mask;
-  const uint8_t prev_byte = matchfinder[1];
-  const uint8_t cur_byte = matchfinder[0];
-  const uint8_t match_byte = matchfinder[reps[0]+1];
+  const int pos_state = data_position() & pos_state_mask;
+  const uint8_t prev_byte = peek( 1 );
+  const uint8_t cur_byte = peek( 0 );
+  const uint8_t match_byte = peek( reps[0] + 1 );
 
-  trials[0].state = state;
-  trials[1].dis = -1;					// literal
   trials[1].price = price0( bm_match[state()][pos_state] );
   if( state.is_char() )
     trials[1].price += price_literal( prev_byte, cur_byte );
   else
     trials[1].price += price_matched( prev_byte, cur_byte, match_byte );
+  trials[1].dis4 = -1;					// literal
 
   const int match_price = price1( bm_match[state()][pos_state] );
   const int rep_match_price = match_price + price1( bm_rep[state()] );
@@ -362,16 +231,15 @@ int LZ_encoder::sequence_optimizer( const int reps[num_rep_distances],
 
   if( num_trials < min_match_len )
     {
-    trials[0].dis = trials[1].dis;
     trials[0].price = 1;
-    matchfinder.move_pos();
+    trials[0].dis4 = trials[1].dis4;
+    move_pos();
     return 1;
     }
 
+  trials[0].state = state;
   for( int i = 0; i < num_rep_distances; ++i )
     trials[0].reps[i] = reps[i];
-  trials[1].prev_index = 0;
-  trials[1].prev_index2 = single_step_trial;
 
   for( int len = min_match_len; len <= num_trials; ++len )
     trials[len].price = infinite_price;
@@ -402,7 +270,7 @@ int LZ_encoder::sequence_optimizer( const int reps[num_rep_distances],
   int cur = 0;
   while( true )				// price optimization loop
     {
-    matchfinder.move_pos();
+    move_pos();
     if( ++cur >= num_trials )		// no more initialized trials
       {
       backward( cur );
@@ -411,7 +279,7 @@ int LZ_encoder::sequence_optimizer( const int reps[num_rep_distances],
 
     const int num_pairs = read_match_distances();
     const int newlen = ( num_pairs > 0 ) ? pairs[num_pairs-1].len : 0;
-    if( newlen >= matchfinder.match_len_limit() )
+    if( newlen >= match_len_limit )
       {
       pending_num_pairs = num_pairs;
       backward( cur );
@@ -422,7 +290,7 @@ int LZ_encoder::sequence_optimizer( const int reps[num_rep_distances],
     Trial & cur_trial = trials[cur];
     State cur_state;
     {
-    int dis = cur_trial.dis;
+    const int dis4 = cur_trial.dis4;
     int prev_index = cur_trial.prev_index;
     const int prev_index2 = cur_trial.prev_index2;
 
@@ -431,38 +299,30 @@ int LZ_encoder::sequence_optimizer( const int reps[num_rep_distances],
       cur_state = trials[prev_index].state;
       if( prev_index + 1 == cur )			// len == 1
         {
-        if( dis == 0 ) cur_state.set_short_rep();
+        if( dis4 == 0 ) cur_state.set_short_rep();
         else cur_state.set_char();			// literal
         }
-      else if( dis < num_rep_distances ) cur_state.set_rep();
+      else if( dis4 < num_rep_distances ) cur_state.set_rep();
       else cur_state.set_match();
       }
-    else if( prev_index2 == dual_step_trial )		// dis == 0
+    else
       {
-      --prev_index;
-      cur_state = trials[prev_index].state;
-      cur_state.set_char();
-      cur_state.set_rep();
-      }
-    else	// if( prev_index2 >= 0 )
-      {
-      prev_index = prev_index2;
-      cur_state = trials[prev_index].state;
-      if( dis < num_rep_distances ) cur_state.set_rep();
-      else cur_state.set_match();
-      cur_state.set_char();
-      cur_state.set_rep();
+      if( prev_index2 == dual_step_trial )	// dis4 == 0 (rep0)
+        --prev_index;
+      else					// prev_index2 >= 0
+        prev_index = prev_index2;
+      cur_state.set_char_rep();
       }
     cur_trial.state = cur_state;
     for( int i = 0; i < num_rep_distances; ++i )
       cur_trial.reps[i] = trials[prev_index].reps[i];
-    mtf_reps( dis, cur_trial.reps );
+    mtf_reps( dis4, cur_trial.reps );		// literal is ignored
     }
 
-    const int pos_state = matchfinder.data_position() & pos_state_mask;
-    const uint8_t prev_byte = matchfinder[1];
-    const uint8_t cur_byte = matchfinder[0];
-    const uint8_t match_byte = matchfinder[cur_trial.reps[0]+1];
+    const int pos_state = data_position() & pos_state_mask;
+    const uint8_t prev_byte = peek( 1 );
+    const uint8_t cur_byte = peek( 0 );
+    const uint8_t match_byte = peek( cur_trial.reps[0] + 1 );
 
     int next_price = cur_trial.price +
                      price0( bm_match[cur_state()][pos_state] );
@@ -479,33 +339,30 @@ int LZ_encoder::sequence_optimizer( const int reps[num_rep_distances],
     const int match_price = cur_trial.price + price1( bm_match[cur_state()][pos_state] );
     const int rep_match_price = match_price + price1( bm_rep[cur_state()] );
 
-    if( match_byte == cur_byte && next_trial.dis != 0 &&
+    if( match_byte == cur_byte && next_trial.dis4 != 0 &&
         next_trial.prev_index2 == single_step_trial )
       {
-      const int price = rep_match_price +
-                        price_shortrep( cur_state, pos_state );
+      const int price = rep_match_price + price_shortrep( cur_state, pos_state );
       if( price <= next_trial.price )
         {
         next_trial.price = price;
-        next_trial.dis = 0;
+        next_trial.dis4 = 0;				// rep0
         next_trial.prev_index = cur;
         }
       }
 
-    const int available_bytes = std::min( matchfinder.available_bytes(),
-                                          max_num_trials - 1 - cur );
-    if( available_bytes < min_match_len ) continue;
+    const int triable_bytes =
+      std::min( available_bytes(), max_num_trials - 1 - cur );
+    if( triable_bytes < min_match_len ) continue;
 
-    const int len_limit = std::min( matchfinder.match_len_limit(),
-                                    available_bytes );
+    const int len_limit = std::min( match_len_limit, triable_bytes );
 
     // try literal + rep0
     if( match_byte != cur_byte && next_trial.prev_index != cur )
       {
-      const uint8_t * const data = matchfinder.ptr_to_current_pos();
+      const uint8_t * const data = ptr_to_current_pos();
       const int dis = cur_trial.reps[0] + 1;
-      const int limit = std::min( matchfinder.match_len_limit() + 1,
-                                  available_bytes );
+      const int limit = std::min( match_len_limit + 1, triable_bytes );
       int len = 1;
       while( len < limit && data[len-dis] == data[len] ) ++len;
       if( --len >= min_match_len )
@@ -513,9 +370,9 @@ int LZ_encoder::sequence_optimizer( const int reps[num_rep_distances],
         const int pos_state2 = ( pos_state + 1 ) & pos_state_mask;
         State state2 = cur_state; state2.set_char();
         const int price = next_price +
-                  price1( bm_match[state2()][pos_state2] ) +
-                  price1( bm_rep[state2()] ) +
-                  price_rep0_len( len, state2, pos_state2 );
+                          price1( bm_match[state2()][pos_state2] ) +
+                          price1( bm_rep[state2()] ) +
+                          price_rep0_len( len, state2, pos_state2 );
         while( num_trials < cur + 1 + len )
           trials[++num_trials].price = infinite_price;
         trials[cur+1+len].update2( price, cur + 1 );
@@ -527,9 +384,9 @@ int LZ_encoder::sequence_optimizer( const int reps[num_rep_distances],
     // try rep distances
     for( int rep = 0; rep < num_rep_distances; ++rep )
       {
-      const uint8_t * const data = matchfinder.ptr_to_current_pos();
-      int len;
+      const uint8_t * const data = ptr_to_current_pos();
       const int dis = cur_trial.reps[rep] + 1;
+      int len;
 
       if( data[0-dis] != data[0] || data[1-dis] != data[1] ) continue;
       for( len = min_match_len; len < len_limit; ++len )
@@ -545,8 +402,7 @@ int LZ_encoder::sequence_optimizer( const int reps[num_rep_distances],
 
       // try rep + literal + rep0
       int len2 = len + 1;
-      const int limit = std::min( matchfinder.match_len_limit() + len2,
-                                  available_bytes );
+      const int limit = std::min( match_len_limit + len2, triable_bytes );
       while( len2 < limit && data[len2-dis] == data[len2] ) ++len2;
       len2 -= len + 1;
       if( len2 < min_match_len ) continue;
@@ -576,22 +432,20 @@ int LZ_encoder::sequence_optimizer( const int reps[num_rep_distances],
         trials[++num_trials].price = infinite_price;
 
       int i = 0;
-      while( start_len > pairs[i].len ) ++i;
+      while( pairs[i].len < start_len ) ++i;
       int dis = pairs[i].dis;
       for( int len = start_len; ; ++len )
         {
         int price = normal_match_price + price_pair( dis, len, pos_state );
-
         trials[cur+len].update( price, dis + num_rep_distances, cur );
 
         // try match + literal + rep0
         if( len == pairs[i].len )
           {
-          const uint8_t * const data = matchfinder.ptr_to_current_pos();
+          const uint8_t * const data = ptr_to_current_pos();
           const int dis2 = dis + 1;
           int len2 = len + 1;
-          const int limit = std::min( matchfinder.match_len_limit() + len2,
-                                      available_bytes );
+          const int limit = std::min( match_len_limit + len2, triable_bytes );
           while( len2 < limit && data[len2-dis2] == data[len2] ) ++len2;
           len2 -= len + 1;
           if( len2 >= min_match_len )
@@ -623,34 +477,33 @@ int LZ_encoder::sequence_optimizer( const int reps[num_rep_distances],
 bool LZ_encoder::encode_member( const unsigned long long member_size )
   {
   const unsigned long long member_size_limit =
-    member_size - File_trailer::size() - max_marker_size;
-  const bool best = ( matchfinder.match_len_limit() > 12 );
+    member_size - Lzip_trailer::size - max_marker_size;
+  const bool best = ( match_len_limit > 12 );
   const int dis_price_count = best ? 1 : 512;
   const int align_price_count = best ? 1 : dis_align_size;
-  const int price_count = ( matchfinder.match_len_limit() > 36 ) ? 1013 : 4093;
-  int price_counter = 0;
+  const int price_count = ( match_len_limit > 36 ) ? 1013 : 4093;
+  int price_counter = 0;		// counters may decrement below 0
   int dis_price_counter = 0;
   int align_price_counter = 0;
   int reps[num_rep_distances];
   State state;
   for( int i = 0; i < num_rep_distances; ++i ) reps[i] = 0;
 
-  if( matchfinder.data_position() != 0 ||
-      renc.member_position() != File_header::size )
+  if( data_position() != 0 || renc.member_position() != Lzip_header::size )
     return false;				// can be called only once
 
-  if( !matchfinder.finished() )			// encode first byte
+  if( !data_finished() )			// encode first byte
     {
     const uint8_t prev_byte = 0;
-    const uint8_t cur_byte = matchfinder[0];
+    const uint8_t cur_byte = peek( 0 );
     renc.encode_bit( bm_match[state()][0], 0 );
     encode_literal( prev_byte, cur_byte );
     crc32.update_byte( crc_, cur_byte );
-    matchfinder.get_match_pairs();
-    matchfinder.move_pos();
+    get_match_pairs();
+    move_pos();
     }
 
-  while( !matchfinder.finished() )
+  while( !data_finished() )
     {
     if( price_counter <= 0 && pending_num_pairs == 0 )
       {
@@ -668,35 +521,32 @@ bool LZ_encoder::encode_member( const unsigned long long member_size )
       }
 
     int ahead = sequence_optimizer( reps, state );
-    if( ahead <= 0 ) return false;		// can't happen
     price_counter -= ahead;
 
     for( int i = 0; ahead > 0; )
       {
-      const int pos_state =
-        ( matchfinder.data_position() - ahead ) & pos_state_mask;
-      const int dis = trials[i].dis;
+      const int pos_state = ( data_position() - ahead ) & pos_state_mask;
       const int len = trials[i].price;
+      int dis = trials[i].dis4;
 
       bool bit = ( dis < 0 );
       renc.encode_bit( bm_match[state()][pos_state], !bit );
       if( bit )					// literal byte
         {
-        const uint8_t prev_byte = matchfinder[ahead+1];
-        const uint8_t cur_byte = matchfinder[ahead];
+        const uint8_t prev_byte = peek( ahead + 1 );
+        const uint8_t cur_byte = peek( ahead );
         crc32.update_byte( crc_, cur_byte );
-        if( state.is_char() )
+        if( state.is_char_set_char() )
           encode_literal( prev_byte, cur_byte );
         else
           {
-          const uint8_t match_byte = matchfinder[ahead+reps[0]+1];
+          const uint8_t match_byte = peek( ahead + reps[0] + 1 );
           encode_matched( prev_byte, cur_byte, match_byte );
           }
-        state.set_char();
         }
       else					// match or repeated match
         {
-        crc32.update_buf( crc_, matchfinder.ptr_to_current_pos() - ahead, len );
+        crc32.update_buf( crc_, ptr_to_current_pos() - ahead, len );
         mtf_reps( dis, reps );
         bit = ( dis < num_rep_distances );
         renc.encode_bit( bm_rep[state()], bit );
@@ -722,9 +572,9 @@ bool LZ_encoder::encode_member( const unsigned long long member_size )
           }
         else					// match
           {
-          encode_pair( dis - num_rep_distances, len, pos_state );
-          if( get_slot( dis - num_rep_distances ) >= end_dis_model )
-            --align_price_counter;
+          dis -= num_rep_distances;
+          encode_pair( dis, len, pos_state );
+          if( dis >= modeled_distances ) --align_price_counter;
           --dis_price_counter;
           match_len_prices.decrement_counter( pos_state );
           state.set_match();
@@ -733,12 +583,12 @@ bool LZ_encoder::encode_member( const unsigned long long member_size )
       ahead -= len; i += len;
       if( renc.member_position() >= member_size_limit )
         {
-        if( !matchfinder.dec_pos( ahead ) ) return false;
-        full_flush( matchfinder.data_position(), state );
+        if( !dec_pos( ahead ) ) return false;
+        full_flush( state );
         return true;
         }
       }
     }
-  full_flush( matchfinder.data_position(), state );
+  full_flush( state );
   return true;
   }
